//===-- X86MASMPrinter.cpp - Convert X86 LLVM code to MASM assembly -------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file was developed by the LLVM research group and is distributed under
// the University of Illinois Open Source License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// MASM Microsoft Assembler printer.
//
//===----------------------------------------------------------------------===//

#include "X86MASMPrinter.h"

#include "llvm/Module.h"
#include "llvm/Function.h"
#include "llvm/Type.h"
#include "llvm/Constants.h"
#include "llvm/Instructions.h"
#include "llvm/DerivedTypes.h"
#include "llvm/Assembly/Writer.h"
#include "llvm/CodeGen/AsmPrinter.h"
#include "llvm/CodeGen/MachineConstantPool.h"
#include "llvm/Support/Mangler.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/ADT/StringExtras.h"
#include "llvm/Support/CommandLine.h"

using namespace llvm;
using namespace x86;

bool X86MASMPrinter::doInitialization(Module &M) {
  AsmPrinter::doInitialization(M);

  O << "TITLE\t" << M.getModuleIdentifier() << "\n"
    << ".686\n"
    << ".MODEL FLAT\n"
    << "\n";

  O << "VOID STRUC\n"
    << "VOID ENDS\n\n";

  printExternals(M);

  O << "\n";
  //O << "SMART\n";

  return false;
}

void X86MASMPrinter::printExternals( Module &M)
{
  // External variable declarations...

  if (!M.global_empty()) {
    O << "; External variable declarations...\n\n";

    for (Module::global_iterator I = M.global_begin(),
            E = M.global_end(); I != E; ++I)
      if (I->hasExternalLinkage() && !I->hasInitializer()) {
          O << "EXTERN " + Mang->getValueName(I) + " : "
            << getMASMTypeName(I->getType()->getElementType())
            << "\n";
      }
    O << "\n";
  }

// External functions...

  if (!M.empty()) {
    O << "; External function declarations...\n\n";
    for (Module::iterator F = M.begin(), E = M.end(); F != E; ++F)
      if ( !F->getIntrinsicID() && F->isExternal())
        O << "EXTERNDEF " << Mang->getValueName(F) << " : NEAR" << "\n";
  }

  O << "EXTERNDEF _malloc : NEAR\n";
  O << "EXTERNDEF _free : NEAR\n";
}

std::string X86MASMPrinter::getMASMTypeName(const Type *Ty) {
  if (Ty->isPrimitiveType())
    switch (Ty->getTypeID()) {
    case Type::VoidTyID:   return "VOID";
    case Type::BoolTyID:   return "BYTE";
    case Type::UByteTyID:  return "CHAR";
    case Type::SByteTyID:  return "SCHAR";
    case Type::UShortTyID: return "WORD";
    case Type::ShortTyID:  return "SWORD";
    case Type::UIntTyID:   return "DWORD";
    case Type::IntTyID:    return "SDWORD";
    case Type::ULongTyID:  return "QWORD";
    case Type::LongTyID:   return "SQWORD";
    case Type::FloatTyID:  return "REAL4";
    case Type::DoubleTyID: return "REAL8";
    default :
      std::cerr << "Unknown primitive type: " << *Ty << "\n";
      abort();
    }

  switch (Ty->getTypeID()) {
  case Type::FunctionTyID:
    return "NEAR";
  case Type::StructTyID: {
    const StructType *STy = cast<StructType>(Ty);
    return "VOID";
  }
  case Type::PointerTyID: {
    const PointerType *PTy = cast<PointerType>(Ty);
    return "PTR " + getMASMTypeName(PTy->getElementType());
  }
  case Type::ArrayTyID:
  {
    const ArrayType *ATy = cast<ArrayType>(Ty);
    return "PTR " + getMASMTypeName(ATy->getElementType());
  }
  case Type::OpaqueTyID:
    return "VOID";
  default:
    assert(0 && "Unhandled case in getTypeProps!");
    abort();
  }

  return "";
}

/// runOnMachineFunction - This uses the printMachineInstruction()
/// method to print assembly for each instruction.
///
bool X86MASMPrinter::runOnMachineFunction(MachineFunction &MF) {
  setupMachineFunction(MF);

  // Print out constants referenced by the function
  printConstantPool(MF.getConstantPool());

  O << "; implementation...\n\n";

  // Print out labels for the function.
  O << "\t.CODE\n";
  emitAlignment(4);
  O << "PUBLIC\t" << CurrentFnName << "\n";
  O << CurrentFnName << "\tPROC NEAR\n";

  // Print out code for the function.
  for (MachineFunction::const_iterator I = MF.begin(), E = MF.end();
       I != E; ++I) {
    // Print a label for the basic block if there are any predecessors.
    if (I->pred_begin() != I->pred_end())
      O << "@@" << CurrentFnName << "_" << I->getNumber() << ":\n";
    for (MachineBasicBlock::const_iterator II = I->begin(), E = I->end();
         II != E; ++II) {
      // Print the assembly for the instruction.
      O << "\t";
      printMachineInstruction(II);
    }
  }

  O << CurrentFnName << "\tENDP\n";

  // We didn't modify anything.
  return false;
}

/// printMachineInstruction -- Print out a single X86 LLVM instruction
/// MI in Intel syntax to the current output stream.
///
void X86MASMPrinter::printMachineInstruction(const MachineInstr *MI) {
  ++EmittedInsts;

  // Call the autogenerated instruction printer routines.
  printInstruction(MI);
}

void X86MASMPrinter::printOperand(const MachineInstr *MI, unsigned OpNo, MVT::ValueType VT){
  const MachineOperand &MO = MI->getOperand(OpNo);
  if (MO.getType() == MachineOperand::MO_MachineRegister) {
    assert(MRegisterInfo::isPhysicalRegister(MO.getReg())&&"Not physref??");
    // Bug Workaround: See note in Printer::doInitialization about %.
    std::string regName = TM.getRegisterInfo()->get(MO.getReg()).Name;
    O << regName;
  } else {
    printOp(MO);
  }
}

void X86MASMPrinter::printOp(const MachineOperand &MO,
                                 bool elideOffsetKeyword /* = false */) {
  const MRegisterInfo &RI = *TM.getRegisterInfo();
  switch (MO.getType()) {
  case MachineOperand::MO_VirtualRegister:
    if (Value *V = MO.getVRegValueOrNull()) {
      O << "<" << V->getName() << ">";
      return;
    }
    // FALLTHROUGH
  case MachineOperand::MO_MachineRegister:
    if (MRegisterInfo::isPhysicalRegister(MO.getReg()))
      // Bug Workaround: See note in Printer::doInitialization about %.
      O << RI.get(MO.getReg()).Name;
    else
      O << "%reg" << MO.getReg();
    return;

  case MachineOperand::MO_SignExtendedImmed:
  case MachineOperand::MO_UnextendedImmed:
    O << (int)MO.getImmedValue();
    return;
  case MachineOperand::MO_MachineBasicBlock: {
    MachineBasicBlock *MBBOp = MO.getMachineBasicBlock();
    O << "@@" << Mang->getValueName(MBBOp->getParent()->getFunction())
      << "_" << MBBOp->getNumber ();
    // O << "\t" << CommentString << " " << MBBOp->getBasicBlock()->getName();
    return;
  }
  case MachineOperand::MO_PCRelativeDisp:
    std::cerr << "Shouldn't use addPCDisp() when building X86 MachineInstrs";
    abort ();
    return;
  case MachineOperand::MO_GlobalAddress: {
    if (!elideOffsetKeyword)
      O << "OFFSET ";
    O << Mang->getValueName(MO.getGlobal());
    int Offset = MO.getOffset();
    if (Offset > 0)
      O << " + " << Offset;
    else if (Offset < 0)
      O << " - " << -Offset;
    return;
  }
  case MachineOperand::MO_ExternalSymbol:
    O << GlobalPrefix << MO.getSymbolName();
    return;
  default:
    O << "<unknown operand type>"; return;
  }
}

void X86MASMPrinter::printCallOperand(const MachineInstr *MI, unsigned OpNo,
                      MVT::ValueType VT) {
  printOp(MI->getOperand(OpNo), true); // Don't print "OFFSET".
}

void X86MASMPrinter::printMemoryOperand(const MachineInstr *MI, unsigned OpNo,
                        MVT::ValueType VT) {
  switch (VT) {
  default: assert(0 && "Unknown arg size!");
  case MVT::i8:   O << "BYTE PTR "; break;
  case MVT::i16:  O << "WORD PTR "; break;
  case MVT::i32:
  case MVT::f32:  O << "DWORD PTR "; break;
  case MVT::i64:
  case MVT::f64:  O << "QWORD PTR "; break;
  case MVT::f80:  O << "TBYTE PTR "; break;
  }
  printMemReference(MI, OpNo);
}

void X86MASMPrinter::printMemReference(const MachineInstr *MI, unsigned Op){
  assert(isMem(MI, Op) && "Invalid memory reference!");

  const MachineOperand &BaseReg  = MI->getOperand(Op);
  int ScaleVal                   = MI->getOperand(Op+1).getImmedValue();
  const MachineOperand &IndexReg = MI->getOperand(Op+2);
  const MachineOperand &DispSpec = MI->getOperand(Op+3);

  if (BaseReg.isFrameIndex()) {
    O << "[frame slot #" << BaseReg.getFrameIndex();
    if (DispSpec.getImmedValue())
      O << " + " << DispSpec.getImmedValue();
    O << "]";
    return;
  } else if (BaseReg.isConstantPoolIndex()) {
    O << "[.CPI" << CurrentFnName << "_"
      << BaseReg.getConstantPoolIndex();

    if (IndexReg.getReg()) {
      O << " + ";
      if (ScaleVal != 1)
        O << ScaleVal << "*";
      printOp(IndexReg);
    }

    if (DispSpec.getImmedValue())
      O << " + " << DispSpec.getImmedValue();
    O << "]";
    return;
  }

  O << "[";
  bool NeedPlus = false;
  if (BaseReg.getReg()) {
    printOp(BaseReg, true);
    NeedPlus = true;
  }

  if (IndexReg.getReg()) {
    if (NeedPlus) O << " + ";
    if (ScaleVal != 1)
      O << ScaleVal << "*";
    printOp(IndexReg);
    NeedPlus = true;
  }

  if (DispSpec.isGlobalAddress()) {
    if (NeedPlus)
      O << " + ";
    printOp(DispSpec, true);
  } else {
    int DispVal = DispSpec.getImmedValue();
    if (DispVal || (!BaseReg.getReg() && !IndexReg.getReg())) {
      if (NeedPlus)
        if (DispVal > 0)
          O << " + ";
        else {
          O << " - ";
          DispVal = -DispVal;
        }
      O << DispVal;
    }
  }
  O << "]";
}

bool X86MASMPrinter::doFinalization(Module &M) {
  const TargetData &TD = TM.getTargetData();
  std::string CurSection;

  O << "\n\n";

  // Print out module-level global variables here.
  for (Module::const_global_iterator I = M.global_begin(), E = M.global_end(); I != E; ++I)
    if (I->hasInitializer()) {
      std::string name = Mang->getValueName(I);
      Constant *C = I->getInitializer();
      unsigned Size = TD.getTypeSize(C->getType());
      unsigned Align = TD.getTypeAlignmentShift(C->getType());

      if (C->isNullValue() &&
          (I->hasLinkOnceLinkage() || I->hasInternalLinkage() ||
           I->hasWeakLinkage() /* FIXME: Verify correct */)) {
        SwitchSection(O, CurSection, ".DATA");
        O << "\tCOMM NEAR " << name << ":" << getMASMTypeName(C->getType());

        O << "\t\t; ";
        WriteAsOperand(O, I, true, true, &M);
        O << "\n";

      } else {
        switch (I->getLinkage()) {
        case GlobalValue::LinkOnceLinkage:
        case GlobalValue::WeakLinkage:   // FIXME: Verify correct for weak.
          // Nonnull linkonce -> weak
          SwitchSection(O, CurSection, ".DATA");
          break;
        case GlobalValue::AppendingLinkage:
          // FIXME: appending linkage variables should go into a section of
          // their name or something.  For now, just emit them as external.
        case GlobalValue::ExternalLinkage:
          if (C->isNullValue())
            SwitchSection(O, CurSection, ".DATA?");
          else
            SwitchSection(O, CurSection, ".DATA");
          // If external or appending, declare as a global symbol
          O << "PUBLIC " << name << "\n";
          break;
        case GlobalValue::InternalLinkage:
          if (C->isNullValue())
            SwitchSection(O, CurSection, ".DATA?");
          else
            SwitchSection(O, CurSection, ".DATA");
          break;
        case GlobalValue::GhostLinkage:
          std::cerr << "GhostLinkage cannot appear in X86AsmPrinter!\n";
          abort();
        }

        emitAlignment(Align);
        O << name << "\t";

        if (!C->isNullValue())
          emitGlobalConstant(C);
        else {
          const TargetData &TD = TM.getTargetData();
          O << "db\t" << TD.getTypeSize(C->getType()) << " dup(?)\n";
          }
      }
    };

  O << "END\n";

  return AsmPrinter::doFinalization(M);
}

// Include the auto-generated portion of the assembly writer.
#include "X86GenAsmWriter2.inc"
